from lpd8.programs import Programs

class Pad:
    """
    Class that defines a single pad
    A pad can have multiple modes and these modes may be combined
    """

    NO_MODE = 0     # Doesn't react to user actions
    SWITCH_MODE = 1 # Switches between 1 and 0 values, both sent at NOTE ON and NOTE OFF events
    PUSH_MODE = 2   # Always sends a 1 at Note ON event and a 0 at NOTE OFF event
    PAD_MODE = 4    # Default mode, acts as a normal pad (sends note value and velocity)
    BLINK_MODE = 8  # May be combined with above modes. Blinks pad at each pad_update call

    OFF = 0
    ON = 1
    BLINK = 2

    def __init__(self, mode=PAD_MODE):
        self.set_mode(mode)
        self._state = self.OFF

    def get_state(self):
        """
        According to the working mode of the pad, returns appropriate value
        :return: The state value
        """
        state = self.OFF
        if self.get_mode(without_blink_mode=False) >= self.BLINK_MODE:
            if self._state == self.ON:
                state = self.ON
            else:
                state = self.BLINK
        elif self._state == self.ON:
            state = self.ON
        return state

    def get_mode(self, without_blink_mode=True):
        """
        Get defined action for this pad. We need this method to get only the action without the blink mode
        :return: mode value without BLINK mode
        """
        if self._mode > self.BLINK_MODE and without_blink_mode:
            return self._mode - self.BLINK_MODE
        else:
            return self._mode

    def set_mode(self, mode):
        """
        Sets pad mode
        :param mode: The desired mode - blink mode may be combined with all others
        """
        self._mode = mode

    def set_switch_state(self, state):
        if self.get_mode() == self.SWITCH_MODE and (state == self.OFF or state == self.ON):
            self._state = state
            return True
        else:
            return False

    def note_on(self, velocity):
        mode = self.get_mode()
        if mode == self.SWITCH_MODE:
            if self._state == self.ON:
                self._state = self.OFF
            else:
                self._state = self.ON
            return self._state
        elif mode == self.PUSH_MODE:
            self._state = self.ON
            return self._state
        elif mode == self.PAD_MODE:
            self._state = self.ON
            return velocity
        else:
            return None

    def note_off(self):
        mode = self.get_mode()
        if mode == self.SWITCH_MODE:
            return self._state
        elif mode == self.PUSH_MODE or mode == self.PAD_MODE:
            self._state = self.OFF
            return self._state
        else:
            return None


class Pads:
    """
    Class that defines a full array of pads (8 pads in each program so 4 X 8 = 32 pads in total
    """

    PAD_1 = 60
    PAD_2 = 62
    PAD_3 = 64
    PAD_4 = 65
    PAD_5 = 67
    PAD_6 = 69
    PAD_7 = 71
    PAD_8 = 72

    ALL_PADS = [PAD_1, PAD_2, PAD_3, PAD_4, PAD_5, PAD_6, PAD_7, PAD_8]
    PAD_MAX = len(ALL_PADS)

    _pad_index = {
        PAD_1: 1,
        PAD_2: 2,
        PAD_3: 3,
        PAD_4: 4,
        PAD_5: 5,
        PAD_6: 6,
        PAD_7: 7,
        PAD_8: 8
    }

    def __init__(self, programs=Programs.PGM_MAX, pads=PAD_MAX):
        self._pads = []
        for program in range(programs + 1):
            self._pads.append([])
            for pad in range(pads + 1):
                self._pads[program].append(Pad())

    def get_mode(self, program, pad):
        return self._pads[program][self._pad_index[pad]].get_mode()

    def set_mode(self, program, pad, mode):
        self._pads[program][self._pad_index[pad]].set_mode(mode)

    def note_on(self, program, pad, velocity):
        return self._pads[program][self._pad_index[pad]].note_on(velocity)

    def note_off(self, program, pad):
        return self._pads[program][self._pad_index[pad]].note_off()

    def set_switch_state(self, program, pad, state):
        return self._pads[program][self._pad_index[pad]].set_switch_state(state)

    def get_state(self, program, pad):
        return self._pads[program][self._pad_index[pad]].get_state()



import rtmidi
from threading import Thread, Timer
from time import sleep
from lpd8.dispatcher import Dispatcher
from lpd8.programs import Programs
from lpd8.knobs import Knobs
from lpd8.pads import Pad, Pads

class LPD8(Thread):
    """
    Main class defining the LPD8 object
    """

    NAME = 'LPD8'
    DELAY = 1
    NOTE_ON = 144
    NOTE_OFF = 128
    CTRL = 176
    PGM_CHG = 192
    BLINK = 100

    def __init__(self, program=4):
        self._program = program
        self._delay = self.DELAY / 1000
        self._blink = self.BLINK / 1000
        self._dispatcher = Dispatcher()
        self._pads = Pads()
        self._knobs = Knobs()
        self.connect()

    def _get_midi_device(self, devices_list):
        midi_device = None
        midi_ports = devices_list.get_ports()
        if len(midi_ports) != 0:
            index = 0
            for name in midi_ports:
                if name.find(self.NAME) != -1:
                    devices_list.open_port(index)
                    midi_device = devices_list
                index += 1
        return midi_device

    def _read_midi(self):
        msg = self._midi_in.get_message()
        if msg is not None:
            cmd = msg[0][0]
            ctrl = msg[0][1]

            if cmd <= self.NOTE_OFF + Programs.PGM_MAX:
                pad_value = self._pads.note_off(cmd - self.NOTE_OFF + 1, ctrl)
                if pad_value is not None:
                    if pad_value == Pad.ON:
                        self.pad_on(cmd - self.NOTE_OFF + 1, [ctrl])
                    self._dispatcher.notify(cmd - self.NOTE_OFF + 1, self.NOTE_OFF, ctrl, pad_value)

            elif cmd <= self.NOTE_ON + Programs.PGM_MAX:
                pad_value = self._pads.note_on(cmd - self.NOTE_ON + 1, ctrl, msg[0][2])
                if pad_value is not None:
                    self._dispatcher.notify(cmd - self.NOTE_ON + 1, self.NOTE_ON, ctrl, pad_value)

            elif cmd <= self.CTRL + Programs.PGM_MAX:
                if ctrl <= Knobs.KNOB_MAX:
                    knob_value = self._knobs.get_value(cmd - self.CTRL + 1, ctrl, msg[0][2])
                    if knob_value is not None:
                        self._dispatcher.notify(cmd - self.CTRL + 1, self.CTRL, ctrl, knob_value)

            elif cmd <= self.PGM_CHG + Programs.PGM_MAX:
                self._dispatcher.notify(cmd - self.PGM_CHG + 1, self.PGM_CHG, ctrl + 1)

    def run(self):
        while self._running:
            self._read_midi()
            sleep(self._delay)

    def stop(self):
        self._running = False

    def is_running(self):
        return self._running

    def connect(self):
        self._midi_in = self._get_midi_device(rtmidi.MidiIn())
        self._midi_out = self._get_midi_device(rtmidi.MidiOut())
        if self._midi_in is not None and self._midi_out is not None:
            self._running = True
        else:
            self._running = False
            print("*** No LPD8 Controller found ***")
        Thread.__init__(self)

    def subscribe(self, callback_method, program, event_type, object_ids):
        if isinstance(object_ids, list):
            for object_id in object_ids:
                self._dispatcher.subscribe(callback_method, program, event_type, object_id)
        else:
            self._dispatcher.subscribe(callback_method, program, event_type, object_ids)

    def unsubscribe(self, program, event_type, object_id=None):
        self._dispatcher.unsubscribe(program, event_type, object_id)

    def set_knob_limits(self, program, knobs, min_value, max_value, is_int=True, is_exp=False, steps=0):
        if isinstance(knobs, list):
            for knob in knobs:
                self._knobs.set_limits(program, knob, min_value, max_value, is_int, is_exp, steps)
        else:
            self._knobs.set_limits(program, knobs, min_value, max_value, is_int, is_exp, steps)

    def set_knob_value(self, program, knobs, value):
        if isinstance(knobs, list):
            for knob in knobs:
                self._knobs.set_value(program, knob, value)
        else:
            self._knobs.set_value(program, knobs, value)

    def set_pad_mode(self, program, pads, mode):
        if isinstance(pads, list):
            for pad in pads:
                self._pads.set_mode(program, pad, mode)
        else:
            self._pads.set_mode(program, pads, mode)

    def set_sticky_knob(self, program, knobs):
        if isinstance(knobs, list):
            for knob in knobs:
                self._knobs.set_sticky(program, knob)
        else:
            self._knobs.set_sticky(program, knobs)

    def set_not_sticky_knob(self, program, knobs):
        if isinstance(knobs, list):
            for knob in knobs:
                self._knobs.set_not_sticky(program, knob)
        else:
            self._knobs.set_not_sticky(program, knobs)

    def set_pad_switch_state(self, program, pads, state):
        if isinstance(pads, list):
            for pad in pads:
                self._pads.set_switch_state(program, pad, state)
        else:
            self._pads.set_switch_state(program, pads, state)

    def pad_update(self):
        if  self._running:
            on_array = []
            off_array = []
            blink_array = []
            for pad in Pads.ALL_PADS:
                state = self._pads.get_state(self._program, pad)
                if state == Pad.BLINK:
                    blink_array.append(pad)
                elif state != Pad.OFF:
                    on_array.append(pad)
                else:
                    off_array.append(pad)
            self.pad_on(self._program, on_array + blink_array)
            self.pad_off(self._program, off_array)
            blink_timer = Timer(self._blink, self.pad_off, [self._program, blink_array])
            blink_timer.start()
            return True
        else:
            return False

    def pad_on(self, program, pads):
        if self._running:
            for pad in pads:
                note_on = [self.NOTE_ON + program - 1, pad, 1]
                self._midi_out.send_message(note_on)
            return True
        else:
            return False

    def pad_off(self, program, pads):
        if self._running:
            for pad in pads:
                note_off = [self.NOTE_ON + program - 1, pad, 0]
                self._midi_out.send_message(note_off)
            return True
        else:
            return False



class Subscriber:
    """
    Class that defines a subscriber to a specific event
    """

    def __init__(self, callback_method, program, event_type, object_id):
        self._callback_method = callback_method
        self._program = program
        self._event_type = event_type
        self._object_id = object_id

    def match(self, program, event_type, object_id):
        """
        Looks for a subscriber to a specific event
        :param program: The program as defined in Program class
        :param event_type: The event type (NOTE_ON, NOTE_OFF, CTRL or PGM_CHG)
        :param object_id: The knob, pad or program change ID
        """
        match = False
        if self._program == program and self._event_type == event_type and self._object_id == object_id:
            match = True
        return match

    def notify(self, program, object_id, data):
        """
        Call the subscriber (subscribed method of an object) and send available data if any
        :param program: The program as defined in Program class
        :param object_id: The knob, pad or program change ID
        :param data: The data to send to the subscriber for processing if any
        """
        if data == None:
            self._callback_method([program, object_id])
        else:
            self._callback_method([program, object_id, data])



import math
from math import floor
from lpd8.programs import Programs

class Knob:
    """
    Class that defines a single control knob
    """

    _MIDI_STEPS = 127   # Real number of MIDI steps
    _TOLERANCE = 2      # Used to compute sticky effect when changing program

    def __init__(self):
        self._value = None
        self._midi_value = None
        self._sticky = True
        self._sync = 0
        self._exp_coeff = [0, 0]
        self.set_limits()

    # Adjusts value if it is an integer value or if steps are defined
    def _adjust_value(self, value):
        if self._inc != 0:
            value = floor(value / self._inc) * self._inc
        if self._is_int:
            value = int(value)
        else:
            value = round(value, 2)
        return value

    def get_value(self, midi_value):
        """
        Gets knob real value according to its properties
        :param midi_value: The knob read midi value
        :return: The computed real value according to knob properties
        """
        if self._sticky and not self._midi_value == None:
            gap = self._midi_value - midi_value
            if abs(gap) > self._TOLERANCE and self._sync == 0:
                self._sync = gap
            if self._sync < 0 and midi_value <= self._midi_value or self._sync > 0 and midi_value >= self._midi_value:
                self._sync = 0
        if self._sync == 0:
            self._midi_value = midi_value
            if self._is_exp:
                value = self._exp_coeff[0] + self._exp_coeff[1] * (10 ** (midi_value / 100))
            else:
                value = self._min_value + (self._max_value - self._min_value) * midi_value / self._MIDI_STEPS
            value = self._adjust_value(value)
            if value == self._value:
                return None
            else:
                self._value = value
                return value

    def set_value(self, value):
        """
        Sets a knob real value, evaluates its associated midi value and stores it
        :param value: The knob real value
        :return: True if value is between minimum and maximum range, false otherwise
        """
        if value >= self._min_value and value <= self._max_value:
            if self._is_exp:
                self._midi_value = int(100 * math.log10((value - self._exp_coeff[0]) / self._exp_coeff[1]))
            else:
                self._midi_value = int(self._MIDI_STEPS * (value - self._min_value) / (self._max_value - self._min_value))
            self._value = self._adjust_value(value)
            return True
        else:
            return False

    def set_sticky(self):
        """
        Sets sticky knob (memory between different programs)
        """
        self._sticky = True

    def set_not_sticky(self):
        """
        Sets not sticky knob (no memory between different programs)
        """
        self._sticky = False

    def set_limits(self, min_value=0, max_value=_MIDI_STEPS, is_int=True, is_exp=False, steps=0):
        """
        Sets knob limits and behaviour
        :param min_value: The minimum value (when MIDI value is equal to 0)
        :param max_value: The maximum value (when MIDI value is equal to 127)
        :param is_int: If true, knob will only return integer values (default), otherwise float values
        :param is_exp: If true, knob values will follow an exponential curve instead of a linear one
        :param steps: Number of steps between min and max values (O to 100 with 10 steps gives 0, 10, ... increments)
        """
        self._min_value = min_value
        self._max_value = max_value
        self._is_int = is_int
        self._is_exp = is_exp
        if is_exp:
            self._exp_coeff[1] = (max_value - min_value) / (10 ** (self._MIDI_STEPS / 100) - 1)
            self._exp_coeff[0] = min_value - self._exp_coeff[1]
        self._inc = 0
        if steps != 0:
            self._inc = (max_value - min_value) / steps

class Knobs:
    """
    Class that defines a full array of knobs (8 knobs in each program so 4 X 8 = 32 knobs in total
    """

    KNOB_1 = 1
    KNOB_2 = 2
    KNOB_3 = 3
    KNOB_4 = 4
    KNOB_5 = 5
    KNOB_6 = 6
    KNOB_7 = 7
    KNOB_8 = 8

    ALL_KNOBS = [KNOB_1, KNOB_2, KNOB_3, KNOB_4, KNOB_5, KNOB_6, KNOB_7, KNOB_8]
    KNOB_MAX = len(ALL_KNOBS)

    def __init__(self, programs=Programs.PGM_MAX, knobs=KNOB_MAX):
        self._knobs = []
        for program in range(programs + 1):
            self._knobs.append([])
            for knob in range(knobs + 1):
                self._knobs[program].append(Knob())

    def get_value(self, program, knob, midi_value):
        """
        Get the value of a knob from a knob array
        :param program:
        :param knob:
        :param midi_value:
        :return:
        """
        return self._knobs[program][knob].get_value(midi_value)

    def set_limits(self, program, knob, min_value, max_value, is_int=True, is_exp=False, steps=0):
        """
        Set knob limits and behaviour in a knob array
        :param program:
        :param knob:
        :param min_value: The minimum value (when MIDI value is equal to 0)
        :param max_value: The maximum value (when MIDI value is equal to 127)
        :param is_int: If true, knob will only return integer values (default), otherwise float values
        :param is_exp: If true, knob values will follow an exponential curve instead of a linear one
        :param steps: Number of steps between min and max values (O to 100 with 10 steps gives 0, 10, ... increments)
        :return:
        """
        self._knobs[program][knob].set_limits(min_value, max_value, is_int, is_exp, steps)

    def set_value(self, program, knob, value):
        """
        Set knob value in a knob array
        :param program:
        :param knob:
        :param value:
        :return:
        """
        self._knobs[program][knob].set_value(value)

    def set_sticky(self, program, knob):
        """
        Set sticky knob in a knob array
        :param program:
        :param knob:
        :return:
        """
        self._knobs[program][knob].set_sticky()

    def set_not_sticky(self, program, knob):
        """
        Set not sticky knob in a knob array
        :param program:
        :param knob:
        :return:
        """
        self._knobs[program][knob].set_not_sticky()



from lpd8.subscriber import Subscriber

class Dispatcher:
    """
    Class used to store subscribers to LPD8 events. A subscriber may be added or suppressed from the dispatcher array
    When a specific event arrives, the notify method is used to retrieve all subscribers to this event and send them
    a notification and associated data
    """
    def __init__(self):
        self._subscribers = []

    def subscribe(self, callback_method, program, event_type, object_id):
        """
        Method used to subscribe to a particular event
        :param callback_method: The callback method itself
        :param program: The program as defined in Program class
        :param event_type: The event type (NOTE_ON, NOTE_OFF, CTRL or PGM_CHG)
        :param object_id: The knob, pad or program change ID
        """
        self._subscribers.append(Subscriber(callback_method, program, event_type, object_id))

    def unsubscribe(self, program, event_type, object_id):
        """
        Method used to unsubscribe to a particular event
        :param program: The program as defined in Program class
        :param event_type: The event type (NOTE_ON, NOTE_OFF, CTRL or PGM_CHG)
        :param object_id: The knob, pad or program change ID
        """
        for subscriber in self._subscribers:
            if subscriber.match(program, event_type, object_id):
                self._subscribers.remove(subscriber)
                break

    def notify(self, program, event_type, object_id, data=None):
        """
        Method used to trigger a notification to a specific event if a subscriber exists for this event
        :param program: The program as defined in Program class
        :param event_type: The event type (NOTE_ON, NOTE_OFF, CTRL or PGM_CHG)
        :param object_id: The knob, pad or program change ID
        :param data: The data to be sent to the subscriber
        """
        for subscriber in self._subscribers:
            if subscriber.match(program, event_type, object_id):
                subscriber.notify(program, object_id, data)



class Pgm_Chg:
    """
    A class defining all possible values for a program change event
    Also makes an array containing all program change IDs and defines the maximum number of available program changes
    """

    PGM_CHG_1 = 1
    PGM_CHG_2 = 2
    PGM_CHG_3 = 3
    PGM_CHG_4 = 4
    PGM_CHG_5 = 5
    PGM_CHG_6 = 6
    PGM_CHG_7 = 7
    PGM_CHG_8 = 8

    ALL_PGM_CHG = [PGM_CHG_1, PGM_CHG_2, PGM_CHG_3, PGM_CHG_4, PGM_CHG_5, PGM_CHG_6, PGM_CHG_7, PGM_CHG_8]
    PGM_MAX = len(ALL_PGM_CHG)



class Programs:
    """
    A class defining all existing programs
    Also defines the maximum number of available programs
    """

    PGM_1 = 1
    PGM_2 = 2
    PGM_3 = 3
    PGM_4 = 4

    PGM_MAX = 4



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lpd8",
    version="1.1.1  ",
    author="Christophe Bury",
    author_email="zetof@zetof.net",
    description="A Python library to drive an AKAI LPD8 pad controller",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zetof/LPD8",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'python-rtmidi',
    ],
    python_requires='>=3.6',
)